import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

/**
 * A first-in, first-out bounded collection of Message objects.
 */
public class BoundedQueue
{
    private Message elements[];
    private int head;
    private int tail;
    private int size;
    private boolean debug;
    
    private Lock qLock = new ReentrantLock();
    private Condition spaceAvailable   = qLock.newCondition();
    private Condition messageAvailable = qLock.newCondition();

    /**
     * Constructs an empty queue.
     * @param capacity the maximum capacity of the queue
     */
    public BoundedQueue(int capacity) 
    {
        elements = new Message[capacity];
        head = 0;
        tail = 0;
        size = 0;
        debug = true;
    }
    
    public void setDebug(boolean aDebug) { debug = aDebug; }

    /**
     * Removes the object at the head.
     * @return the object that has been removed from the queue 
     * @precondition !isEmpty()
     */
    public Message remove() throws InterruptedException
    {
        qLock.lock();
        
        try {
            while (isEmpty()) messageAvailable.await();
            Message msg = elements[head];
            
            head++;
            size--;
            if (head == elements.length) head = 0;
            
            if (debug) {
                msg.sequence = -msg.sequence;
                System.out.printf("%2d C %s\n", size, msg);
                msg.sequence = -msg.sequence;
            }

            spaceAvailable.signalAll();
            return msg;
        }
        finally {
            qLock.unlock();
        }
    }

    /**
     * Appends an object at the tail.
     * @param newValue the object to be appended 
     * @precondition !isFull();
     */
    public void add(Message newValue) throws InterruptedException
    {
        qLock.lock();
        
        try {
            while (isFull()) spaceAvailable.await();
            elements[tail] = newValue;
            
            tail++;
            size++;
            if (tail == elements.length) tail = 0;
            
            if (debug) System.out.printf("%2d P %s\n", size, newValue);
            messageAvailable.signalAll();
        }
        finally {
            qLock.unlock();
        }
    }
    
    public int getSize() { return size; }

    public boolean isFull() 
    {
        return size == elements.length;
    }

    public boolean isEmpty() 
    {
        return size == 0;
    }
}
